import matplotlib.pylab as plt

import pynbody

# load the snapshot and set to physical units
s = pynbody.load('testdata/gasoline_ahf/g15784.lr.01024.gz')

# load the halos
h = s.halos()

# center on the largest halo and align the disk
pynbody.analysis.angmom.faceon(h[0])

# convert all units to something reasonable (kpc, Msol etc)
s.physical_units()

# create a profile object for the stars (by default this is a 2D profile)
p = pynbody.analysis.profile.Profile(h[0].s, vmin =.01, max=50)

# make the figure and sub plots
f, axs = plt.subplots(1,2,figsize=(14,6))

# make the plot
axs[0].plot(p['rbins'],p['density'], 'k')
axs[0].semilogy()
axs[0].set_xlabel('R [kpc]')
axs[0].set_ylabel(r'$\Sigma_{\star}$ [M$_{\odot}$ kpc$^{-2}$]')
axs[0].set_title("2D stellar density")

# make a 3D density plot of the dark matter (note ndim=3 in the constructor below)
p = pynbody.analysis.profile.Profile(h[0].d,min=.01,max=50,ndim=3)

axs[1].plot(p['rbins'],p['density'], 'k')
axs[1].semilogy()
axs[1].set_xlabel('R [kpc]')
axs[1].set_ylabel(r'$\rho_{DM}$ [M$_{\odot}$ kpc$^{-3}$]')
axs[1].set_title("3D dark matter density")



"""
Galactic bulge/disk/halo decomposition

"""

import logging

import numpy as np

from .. import array, config, filt, transformation, util
from . import angmom, profile

logger = logging.getLogger('pynbody.analysis.decomp')


def decomp(h, aligned=False, j_disk_min=0.8, j_disk_max=1.1, E_cut=None, j_circ_from_r=False,
           log_interp=False, angmom_size="3 kpc"):
    """Creates an array 'decomp' for star particles in the simulation, with an integer specifying components.

    The possible values of the components are:

    1 -- thin disk

    2 -- halo

    3 -- bulge

    4 -- thick disk

    5 -- pseudo bulge

    This routine is based on an original IDL procedure by Chris Brook.

    Arguments
    ---------

    h : SimSnap
        The simulation snapshot to analyze

    aligned : bool
        If True, the simulation is assumed to be already aligned so that the disk is in the xy plane.
        Otherwise, the simulation is recentered and aligned into the xy plane.

    j_disk_min : float
        The minimum angular momentum as a proportion of the circular angular momentum which a particle must have to be
        part of the 'disk'.

    j_disk_max : float
        The maximum angular momentum as a proportion of the circular angular momentum which a particle can have to be
        part of the 'disk'.

    E_cut : float
        The energy boundary between bulge and spheroid. If None, this is taken to be the median energy of the stars.

    j_circ_from_r : bool
        If True, the maximum angular momentum is determined as a function of radius, rather than as a function of
        orbital energy. Default False (determine as function of energy).

    angmom_size : str
        The size of the disk to use for calculating the angular momentum vector. Default is "3 kpc".

    """

    import scipy.interpolate as interp
    global config

    # Center, eliminate proper motion, rotate so that
    # gas disk is in X-Y plane
    if aligned:
        tx = transformation.NullTransformation(h)
    else:
        tx = angmom.faceon(h, disk_size=angmom_size)

    with tx:

        # Find KE, PE and TE
        ke = h['ke']
        pe = h['phi']

        h['phi'].convert_units(ke.units)  # put PE and TE into same unit system

        te = ke + pe
        h['te'] = te
        te_star = h.star['te']

        te_max = te_star.max()

        # Add an arbitrary offset to the PE to reflect the idea that
        # the group is 'fully bound'.
        te -= te_max
        logger.info("te_max = %.2e" % te_max)

        h['te'] -= te_max

        logger.info("Making disk rotation curve...")

        # Now make a rotation curve for the disk. We'll take everything
        # inside a vertical height of eps*3

        d = h[filt.Disc('1 Mpc', h['eps'].min() * 3)]

        try:

            # attempt to load rotation curve off disk
            r, v_c = np.loadtxt(h.ancestor.filename + ".rot." +
                                str(h.properties["halo_number"]), skiprows=1, unpack=True)

            pro_d = profile.Profile(d, nbins=100, type='log')
            r_me = pro_d["rbins"].in_units("kpc")
            r_x = np.concatenate(([0], r, [r.max() * 2]))
            v_c = np.concatenate(([v_c[0]], v_c, [v_c[-1]]))
            v_c = interp.interp1d(r_x, v_c, bounds_error=False)(r_me)

            logger.info(" - found existing rotation curve on disk, using that")

            v_c = v_c.view(array.SimArray)
            v_c.units = "km s^-1"
            v_c.sim = d

            v_c.convert_units(h['vel'].units)

            pro_d._profiles['v_circ'] = v_c
            pro_d.v_circ_loaded = True

        except Exception:
            pro_d = profile.Profile(d, nbins=100, type='log')  # .D()
            # Nasty hack follows to force the full halo to be used in calculating the
            # gravity (otherwise get incorrect rotation curves)
            pro_d._profiles['v_circ'] = profile.v_circ(pro_d, h)

        pro_phi = pro_d['phi']
        #import pdb; pdb.set_trace()
        # offset the potential as for the te array
        pro_phi -= te_max

        # (will automatically be reflected in E_circ etc)
        # calculating v_circ for j_circ and E_circ is slow

        if j_circ_from_r:
            pro_d.create_particle_array("j_circ", out_sim=h)
            pro_d.create_particle_array("E_circ", out_sim=h)
        else:

            if log_interp:
                j_from_E = interp.interp1d(
                    np.log10(-pro_d['E_circ'].in_units(ke.units))[::-1], np.log10(pro_d['j_circ'])[::-1], bounds_error=False)
                h['j_circ'] = 10 ** j_from_E(np.log10(-h['te']))
            else:
                #            j_from_E  = interp.interp1d(-pro_d['E_circ'][::-1], (pro_d['j_circ'])[::-1], bounds_error=False)
                j_from_E = interp.interp1d(
                    pro_d['E_circ'].in_units(ke.units), pro_d['j_circ'], bounds_error=False)
                h['j_circ'] = j_from_E(h['te'])

            # The next line forces everything close-to-unbound into the
            # spheroid, as per CB's original script ('get rid of weird
            # outputs', it says).
            h['j_circ'][np.where(h['te'] > pro_d['E_circ'].max())] = np.inf

            # There are only a handful of particles falling into the following
            # category:
            h['j_circ'][np.where(h['te'] < pro_d['E_circ'].min())] = pro_d[
                'j_circ'][0]

        h['jz_by_jzcirc'] = h['j'][:, 2] / h['j_circ']
        h_star = h.star

        if 'decomp' not in h_star:
            h_star._create_array('decomp', dtype=int)
        disk = np.where(
            (h_star['jz_by_jzcirc'] > j_disk_min) * (h_star['jz_by_jzcirc'] < j_disk_max))

        h_star['decomp', disk[0]] = 1
        # h_star = h_star[np.where(h_star['decomp']!=1)]

        # Find disk/spheroid angular momentum cut-off to make spheroid
        # rotational velocity exactly zero.

        V = h_star['vcxy']
        JzJcirc = h_star['jz_by_jzcirc']
        te = h_star['te']

        logger.info("Finding spheroid/disk angular momentum boundary...")

        j_crit = util.bisect(0., 5.0,
                             lambda c: np.mean(V[np.where(JzJcirc < c)]))

        logger.info("j_crit = %.2e" % j_crit)

        if j_crit > j_disk_min:
            logger.warning(
                "!! j_crit exceeds j_disk_min. This is usually a sign that something is going wrong (train-wreck galaxy?)")
            logger.warning("!! j_crit will be reset to j_disk_min=%.2e" % j_disk_min)
            j_crit = j_disk_min

        sphere = np.where(h_star['jz_by_jzcirc'] < j_crit)

        if E_cut is None:
            E_cut = np.median(h_star['te'])

        logger.info("E_cut = %.2e" % E_cut)

        halo = np.where((te > E_cut) * (JzJcirc < j_crit))
        bulge = np.where((te <= E_cut) * (JzJcirc < j_crit))
        pbulge = np.where((te <= E_cut) * (JzJcirc > j_crit)
                          * ((JzJcirc < j_disk_min) + (JzJcirc > j_disk_max)))
        thick = np.where((te > E_cut) * (JzJcirc > j_crit)
                         * ((JzJcirc < j_disk_min) + (JzJcirc > j_disk_max)))

        h_star['decomp', halo] = 2
        h_star['decomp', bulge] = 3
        h_star['decomp', thick] = 4
        h_star['decomp', pbulge] = 5

    # Return profile object for informational purposes
    return pro_d



import numpy as np
import numpy.testing as npt
import pytest

import pynbody


@pytest.fixture
def snap():
    f = pynbody.new(1000)
    np.random.seed(1337)
    f['pos'] = np.random.normal(scale=1.0, size=f['pos'].shape)
    f['vel'] = np.random.normal(scale=1.0, size=f['vel'].shape)
    f['mass'] = np.random.uniform(1.0, 10.0, size=f['mass'].shape)
    f['pos'].units = 'kpc'
    f['vel'].units = 'km s^-1'
    f['mass'].units = 'Msol'
    return f

@pytest.fixture(params=[-0.5, 0.0], ids=["centred", "zero-origin"])
def wrapping_snap(request):
    f = pynbody.new(5000)
    min = request.param
    np.random.seed(1337)
    f['pos'] = np.random.uniform(0, 1.0, size=f['pos'].shape) + min
    f['vel'] = np.random.normal(scale=1.0, size=f['vel'].shape)
    f['mass'] = np.random.uniform(1.0, 10.0, size=f['mass'].shape)
    f['pos'].units = 'kpc'
    f['vel'].units = 'km s^-1'
    f['mass'].units = 'Msol'
    f.properties['boxsize'] = 1.0
    return f, min




def test_sphere(snap):
    f = snap
    sp = f[pynbody.filt.Sphere(0.5)]
    assert (sp.get_index_list(f) == np.where(f['r'] < 0.5)[0]).all()
    assert sp['x'].max() < 0.5
    assert sp['x'].min() > -0.5
    assert sp['r'].max() < 0.5

    sp_units = f[pynbody.filt.Sphere('500 pc')]
    assert len(sp_units.intersect(sp)) == len(sp)


def test_empty_sphere():
    snap = pynbody.new(0)
    # This would fail due to the C code seeing an array it couldn't handle (due to zero length)
    sp_empty = snap[pynbody.filt.Sphere(1.0)]
    assert len(sp_empty) == 0

def test_wrapping_sphere(wrapping_snap):
    f, min = wrapping_snap
    # put a sphere at the lower left corner of the box
    sp = f[pynbody.filt.Sphere(0.5, (min, min, min))]

    f['pos']-=np.array([min, min, min]) # put our sphere at the origin and re-wrap
    f.wrap()
    f.wrap()

    sphere_unwrapped = f[pynbody.filt.Sphere(0.5)]
    assert (sphere_unwrapped.get_index_list(f) == sp.get_index_list(f)).all()



def test_passfilters(snap):
    f = snap

    hp = f[pynbody.filt.HighPass('mass', 5)]
    lp = f[pynbody.filt.LowPass('mass', 5)]
    bp = f[pynbody.filt.BandPass('mass', 2, 7)]

    assert len(hp) > 0
    assert len(lp) > 0
    assert len(bp) > 0

    assert len(hp.intersect(lp)) == 0

    assert (hp.get_index_list(f) == np.where(f['mass'] > 5)[0]).all()
    assert (lp.get_index_list(f) == np.where(f['mass'] < 5)[0]).all()
    assert (bp.get_index_list(f) == np.where(
        (f['mass'] > 2) * (f['mass'] < 7))[0]).all()

@pytest.mark.parametrize("with_kdtree", [True, False])
def test_cuboid(snap, with_kdtree):
    f = snap
    x1s = np.random.uniform(-0.5, 0.0, 10)
    x2s = np.random.uniform(0.0, 0.5, 10)
    y1s = np.random.uniform(-0.5, 0.0, 10)
    y2s = np.random.uniform(0.0, 0.5, 10)
    z1s = np.random.uniform(-0.5, 0.0, 10)
    z2s = np.random.uniform(0.0, 0.5, 10)

    if with_kdtree:
        f.build_tree()

    for x1, x2, y1, y2, z1, z2 in zip(x1s, x2s, y1s, y2s, z1s, z2s):
        c = f[pynbody.filt.Cuboid(x1, y1, z1, x2, y2, z2)]
        assert (c.get_index_list(f) == np.where(
            (f['x'] > x1) * (f['x'] < x2) * (f['y'] > y1) * (f['y'] < y2) * (f['z'] > z1) * (f['z'] < z2))[0]).all()

def test_wrapping_cuboid(wrapping_snap):
    snap, origin = wrapping_snap

    # NB because 'wrapping' in pynbody only takes place once, if any boundaries get more than one wrap distance
    # away from the origin, they will not be included in the wrapped box. This is a limitation of the current
    # implementation that has been chosen for efficiency. It is likely to be completely fine for all realistic
    # use cases. But it also means that choosing any combination of box_corner and box_size can result in a failure. The test avoids such cases.
    for box_corner in ((0.5,0.5,0.5), (0.3, 0.5, 0.5), (0.2,-0.2,0.0)):
        for box_size in ((0.7, 0.7, 0.5), (0.9, 0.3, 0.1), (0.2, 0.8, 0.2), (0.2, -0.2, 0.2)):
            # nb the last box_size being 'negative' (-0.2) should be equivalent to +0.8, tested via the %1.0 in the
            # 'unwrapped' box
            subbox = snap[pynbody.filt.Cuboid(origin+box_corner[0], origin+box_corner[1], origin+box_corner[2],
                                              *(origin+c+s for c,s in zip(box_corner, box_size)))]

            snap['pos']-=origin+np.array(box_corner)+0.5 # put our cuboid at the bottom left
            snap.wrap()
            snap.wrap() # wrap twice in case any particles have ended up more than one wrap distance away
            # (the limitation mentioned above in not being aable to move particles more than one wrap distance
            # within the filter evaluation is a separate issue from this)

            subbox_unwrapped = snap[pynbody.filt.Cuboid(-0.5, -0.5, -0.5,
                                                        box_size[0]%1.0-0.5, box_size[1]%1.0-0.5, box_size[2]%1.0-0.5)]

            assert (subbox_unwrapped.get_index_list(snap) == subbox.get_index_list(snap)).all()


def test_logic(snap):
    f = snap

    comp = f[pynbody.filt.BandPass('mass', 2, 7)]

    and_test = f[
        pynbody.filt.HighPass('mass', 2) & pynbody.filt.LowPass('mass', 7)]

    assert and_test == comp

    comp = f[pynbody.filt.LowPass('mass', 2)]
    or_test = f[
        (pynbody.filt.LowPass('mass', 1) | pynbody.filt.BandPass('mass', 1, 2))]

    assert or_test == comp

    comp = f[pynbody.filt.BandPass('mass', 2, 7)]
    not_test = f[~(pynbody.filt.BandPass('mass', 2, 7))]

    assert comp.union(not_test) == f
    assert len(comp.intersect(not_test)) == 0

    assert len(comp)+len(not_test)==len(f)
    assert len(comp)!=0
    assert len(not_test)!=0

def test_family_filter():
    f = pynbody.new(dm=100,gas=100)
    f_dm = f.dm
    f_dm_filter = f[pynbody.filt.FamilyFilter(pynbody.family.dm)]
    f_gas = f.gas
    f_gas_filter = f[pynbody.filt.FamilyFilter(pynbody.family.gas)]
    assert (f_dm.get_index_list(f) == f_dm_filter.get_index_list(f)).all()
    assert (f_gas.get_index_list(f) == f_gas_filter.get_index_list(f)).all()

def test_annulus(snap):
    a = snap[pynbody.filt.Annulus(0.5, 1.0)]
    assert len(a)>100 and len(a)<200
    assert (a.get_index_list(snap) == np.where(
        (snap['r'] > 0.5) * (snap['r'] < 1.0))[0]).all()

def test_disc(snap):
    d = snap[pynbody.filt.Disc(0.8, 0.3)]
    assert len(d)>50 and len(d)<100
    assert (d.get_index_list(snap) == np.where(
        (snap['rxy'] < 0.8) * (abs(snap['z']) < 0.3))[0]).all()

def test_solar_neighbourhood(snap):
    filt = pynbody.filt.SolarNeighborhood(0.5, 1.0, 0.3)
    sn = snap[filt]
    assert len(sn) > 50 and len(sn) < 100
    assert (sn.get_index_list(snap) == np.where(
        (snap['rxy'] > 0.5) * (snap['rxy'] < 1.0) * (abs(snap['z']) < 0.3))[0]).all()

def test_hashing():
    X = {}
    X[pynbody.filt.Sphere('100 kpc')] = 5

    X[pynbody.filt.FamilyFilter(pynbody.family.gas)] = 10
    assert X.get(pynbody.filt.Sphere('100 kpc'), None) == 5
    assert X.get(pynbody.filt.FamilyFilter(pynbody.family.gas),None)==10
    with pytest.raises(KeyError):
        X[pynbody.filt.FamilyFilter(pynbody.family.dm)]



"""Support for the Velociraptor halo finder."""

from __future__ import annotations

import functools
import warnings
from pathlib import Path
from typing import Optional

import h5py
import numpy as np
from numpy.typing import NDArray

from .. import array, units, util
from . import HaloCatalogue, HaloParticleIndices
from .details import number_mapping


class VelociraptorCatalogue(HaloCatalogue):
    """Handles catalogues produced by the velociraptor halo finder.
    """

    @classmethod
    def _catalogue_path(cls, sim) -> Path | None:


        simpath = Path(sim.filename)

        filename = simpath.name

        # from filename snap_nnnn[.m].hdf5 get integer n
        try:
            snapshot_num = int(filename.split('_')[1].split('.')[0])
        except (ValueError, IndexError):
            snapshot_num = None

        possible_paths = [
            simpath.parent,
            simpath.parent / 'VR',
            simpath.parent.parent,
            simpath.parent.parent / 'VR'
        ]

        for basepath in possible_paths:
            if basepath.is_dir():
                for p in basepath.iterdir():
                    if snapshot_num and p.is_dir() and f'{snapshot_num:04d}' in p.name:
                        possible_paths.append(p)
                    if 'catalog_groups.0' in p.name and p.is_file():
                        return basepath / str(p.name)[:-(len('.catalog_groups.0'))]
        return None

    @classmethod
    def _can_load(cls, sim, **kwargs):
        path = cls._catalogue_path(sim)
        if path is None:
            return False

        for suffix in ['.catalog_groups.0', '.catalog_particles.0', '.properties.0']:
            if not (path.with_suffix(suffix).is_file()):
                return False

        return True

    def __init__(self, sim, filename=None, include_unbound=False):
        """Create a new velociraptor catalogue object

        Parameters
        ----------
        sim : pynbody.SimSnap
            The simulation snapshot
        filename : str, optional
            The filename of the velociraptor catalogue. If not specified, the code will try to guess the filename.
            Here, the filename is considered to be the stem; e.g. if your velociraptor files are called
            ``folder/output.properties.0`, ``folder/output.catalog_groups.0``, etc., you should specify
            ``filename='folder/output'``.
        include_unbound : bool, optional
            Whether to include unbound particles in the particle list. Default is False.
        """

        self._include_unbound = include_unbound
        if filename is None:
            self._path = self._catalogue_path(sim)
        else:
            self._path = Path(filename)

        if self._path is None:
            raise OSError("Could not find velociraptor catalogue. Try specifying vr_basename='path/to/output', where the velociraptor outputs are output.properties.0 etc")
        self._grps = h5py.File(str(self._path.with_suffix('.catalog_groups.0')), 'r')
        self._part_ids = h5py.File(str(self._path.with_suffix('.catalog_particles.0')), 'r')
        self._properties_hdf_file = h5py.File(str(self._path.with_suffix('.properties.0')), 'r')

        if include_unbound:
            self._part_ids_unbound = h5py.File(str(self._path.with_suffix('.catalog_particles.unbound.0')), 'r')
        self._props = h5py.File(str(self._path.with_suffix('.properties.0')), 'r')

        assert self._grps['Num_of_files'][0] == 1, "Multi-file catalogues not supported at present"

        self._num_halos = self._grps['Num_of_groups'][0]

        super().__init__(sim, number_mapping.create_halo_number_mapper(self._properties_hdf_file['ID']))

        self._setup_property_keys()
        self._setup_property_units()
        self._init_iord_to_fpos()
        self._calculate_children()

    def _setup_property_keys(self):
        self._property_keys = []
        for k in self._properties_hdf_file.keys():
            if len(self._properties_hdf_file[k]) == self._num_halos:
                self._property_keys.append(k)

    def _setup_property_units(self):
        unitinfo = self._props['UnitInfo'].attrs
        comoving = int(unitinfo['Comoving_or_Physical'])!=0
        length = units.Unit("kpc")*float(unitinfo['Length_unit_to_kpc'])
        if comoving:
            # https://github.com/pelahi/VELOCIraptor-STF/blob/6f4b760ef5043b959a922a8e7ae453fd0a9f988f/src/io.cxx#L1591
            length *= units.Unit("a h^-1")
        mass = units.Unit("Msol")*float(unitinfo['Mass_unit_to_solarmass'])
        vel = units.Unit("km s^-1")*float(unitinfo['Velocity_unit_to_kms'])
        time = length / vel
        # the above is actually a guess - but don't think the Dimension_Time is anyway ever used?
        # if any time dimension is found, a warning will be issued

        self._property_units = []
        dims_attr_names = ("Dimension_Length", "Dimension_Mass", "Dimension_Time", "Dimension_Velocity")
        dims_units = (length, mass, time, vel)
        for k in self._property_keys:
            powers = [util.fractions.Fraction.from_float(float(self._props[k].attrs[d])).limit_denominator()
                      for d in dims_attr_names]
            if powers[2] != 0.0:
                warnings.warn("Time dimension found in property %s, but no time conversion factor is stored in the velociraptor output. Guessing an appropriate conversion." % k)
            final_unit = functools.reduce(lambda x, y: x * y,
                                     [u**p for u, p in zip(dims_units, powers)])
            self._property_units.append(final_unit)



    def _calculate_children(self):
        self._parents = self._grps['Parent_halo_ID'][:]
        _all_children_zero_based = np.arange(self._num_halos, dtype=np.int32)[self._parents != -1]
        self._all_children_ordered_by_parent = self.number_mapper.index_to_number(
                _all_children_zero_based[np.argsort(self._parents[_all_children_zero_based])]
        )

        self._children_start_index = np.searchsorted(self._parents[
                                                         self.number_mapper.number_to_index(self._all_children_ordered_by_parent)
                                                     ],
                                                     self.number_mapper.all_numbers,
                                                     side='left')
        self._children_stop_index = np.concatenate((self._children_start_index[1:],
                                                    np.array([self._num_halos],
                                                             dtype=self._children_start_index.dtype)))

    def get_properties_one_halo(self, halo_number) -> dict:
        i_zerobased = self.number_mapper.number_to_index(halo_number)
        properties = {k: self._props[k][i_zerobased] * u
                      for k, u in zip(self._property_keys, self._property_units)}
        parent = self._parents[i_zerobased]
        children = self._all_children_ordered_by_parent[self._children_start_index[i_zerobased]:self._children_stop_index[i_zerobased]]
        properties.update({'parent': parent, 'children': children})
        return properties

    def get_properties_all_halos(self, with_units=True) -> dict:
        if with_units:
            all_properties_hdf_file = {k: array.SimArray(self._properties_hdf_file[k][:], u)
                              for k, u in zip(self._property_keys, self._property_units)}
        else:
            all_properties_hdf_file = {k: self._properties_hdf_file[k] for k in self._property_keys}

        all_properties_hdf_file.update(
               {'parent': self._parents,
                'children': [self._all_children_ordered_by_parent[start:end]
                             for start, end in zip(self._children_start_index, self._children_stop_index)]}
        )
        return all_properties_hdf_file


    def _get_particle_indices_one_halo(self, halo_number) -> NDArray[int]:
        i_zerobased = self.number_mapper.number_to_index(halo_number)
        ptcl_fpos = self.__get_particle_indices_from_halo_index(i_zerobased, False)

        if self._include_unbound:
            ptcl_fpos_unbound =  self.__get_particle_indices_from_halo_index(i_zerobased, True)
            ptcl_fpos = np.concatenate((ptcl_fpos, ptcl_fpos_unbound))

        return np.sort(ptcl_fpos)

    def _get_all_particle_indices(self) -> HaloParticleIndices | tuple[np.ndarray, np.ndarray]:
        particle_ids_hdf_array = self._part_ids['Particle_IDs']
        offsets = np.concatenate((self._grps['Offset'][:], [particle_ids_hdf_array.shape[0]]),
                                 dtype=np.intp)
        boundaries = np.vstack((offsets[:-1], offsets[1:])).T

        if self._include_unbound:
            num_ids = particle_ids_hdf_array.shape[0] + self._part_ids_unbound['Particle_IDs'].shape[0]
            offsets_unbound = np.concatenate((self._grps['Offset_unbound'][:],
                                              [self._part_ids_unbound['Particle_IDs'].shape[0]]),
                                             dtype=np.intp)
            boundaries_unbound = np.vstack((offsets_unbound[:-1], offsets_unbound[1:])).T
            output_boundaries = boundaries + boundaries_unbound
            particle_ids = np.empty(num_ids, dtype=np.intp)

            for (a,b), (a_unbound, b_unbound), (a_out, b_out) in zip(boundaries, boundaries_unbound, output_boundaries):
                particle_ids[a_out:b_out] = np.sort(self._iord_to_fpos.map_ignoring_order(
                    np.concatenate((self._part_ids['Particle_IDs'][a:b],
                                    self._part_ids_unbound['Particle_IDs'][a_unbound:b_unbound])))
                )
        else:
            output_boundaries = boundaries

            particle_ids = np.empty(particle_ids_hdf_array.shape[0], dtype=np.intp)

            for a,b in boundaries:
                particle_ids[a:b] = np.sort(self._iord_to_fpos.map_ignoring_order(self._part_ids['Particle_IDs'][a:b]))

        return particle_ids, output_boundaries


    def __get_particle_indices_from_halo_index(self, i_zerobased, unbound):
        if unbound:
            grps_hdf_array = self._grps['Offset_unbound']
            particle_ids_hdf_array = self._part_ids_unbound['Particle_IDs']
        else:
            grps_hdf_array = self._grps['Offset']
            particle_ids_hdf_array = self._part_ids['Particle_IDs']

        ptcl_start = grps_hdf_array[i_zerobased]

        if i_zerobased == self._num_halos - 1:
            ptcl_end = particle_ids_hdf_array.shape[0]
        else:
            ptcl_end = grps_hdf_array[i_zerobased + 1]

        ptcl_ids_this_halo = particle_ids_hdf_array[ptcl_start:ptcl_end]
        ptcl_fpos = self._iord_to_fpos.map_ignoring_order(ptcl_ids_this_halo)
        return ptcl_fpos



"""
Supports loading of ASCII snapshots. The format is a simple text file with a header line defining columns.

"""

import os
import pathlib

import numpy as np

from .. import chunk, family, units
from . import SimSnap

_max_buf = 1024 * 512


class AsciiSnap(SimSnap):
    """Supports loading of ASCII snapshots. The format is a simple text file with a header line defining columns.

    For an ASCII snapshot to be identified automatically using :func:`pynbody.load`, the file must have a `.txt`
    extension.

    All particles are assumed to be dark matter.
    """

    def _setup_slices(self, num_lines, take=None):
        disk_family_slice = {family.dm: slice(0,num_lines)}
        self._load_control = chunk.LoadControl(
            disk_family_slice, _max_buf, take)
        self._family_slice = self._load_control.mem_family_slice
        self._num_particles = self._load_control.mem_num_particles

    def __init__(self, filename, take=None, header=None):
        super().__init__()

        num_particles = 0

        self._header = header

        with open(filename) as f:
            for l in f:
                if not header:
                    header = l
                else:
                    num_particles+=1

        self._loadable_keys = header.split()
        self._filename = filename
        self._file_units_system = [units.Unit("G"), units.Unit("kpc"), units.Unit("Msol")]
        self._setup_slices(num_particles,take=take)
        self._decorate()

    def loadable_keys(self, fam=None):
        return self._loadable_keys

    def _load_arrays(self, array_name_list):
        with open(self._filename) as f:

            if not self._header: f.readline()
            rs=[]
            for array_name in array_name_list:
                self._create_array(array_name, ndim=1)

            rs = [self[array_name] for array_name in array_name_list]
            cols = [self._loadable_keys.index(array_name) for array_name in array_name_list]
            ncols = len(self._loadable_keys)

            buf_shape = _max_buf

            b = np.empty(buf_shape)

            for readlen, buf_index, mem_index in self._load_control.iterate([family.dm],[family.dm]):
                b = np.fromfile(f, count=readlen*ncols, sep=' ').reshape((-1,ncols))

                if mem_index is not None:
                    for r,col in zip(rs,cols):
                        r[mem_index] = b[buf_index,col]

    def _load_array(self, array_name, fam=None):

        if fam is not None:
            raise OSError("Arrays only loadable at snapshot, not family level")

        ars = [array_name]

        if array_name not in self._loadable_keys:
            ars =  self._array_name_ND_to_1D(array_name)
            for array_name_i in ars:
                if array_name_i not in self._loadable_keys:
                    raise OSError("No such array on disk")

        self._load_arrays(ars)

    @classmethod
    def _can_load(cls, f: pathlib.Path):
        return f.exists() and f.suffix == '.txt'



import numpy as np
import pytest
from numpy.testing import assert_allclose

import pynbody


@pytest.fixture(params=[True, False])
def snap(request):
    f = pynbody.load("testdata/subfind/snapshot_019")
    if request.param:
        f['iord'] = np.random.permutation(f['iord'])
    return f

@pytest.fixture
def halos(snap):
    return pynbody.halo.subfind.SubfindCatalogue(snap, subhalos=True)

@pytest.fixture
def groups(snap):
    return pynbody.halo.subfind.SubfindCatalogue(snap, subhalos=False)


def test_lengths(halos, groups):
    assert len(groups)==2853
    assert len(halos)==3290

def test_deprecated_subs_keyword(snap, halos):
    with pytest.warns(DeprecationWarning):
        new_halos = snap.halos(subs=True)
        assert isinstance(new_halos, pynbody.halo.subfind.SubfindCatalogue)
        assert len(new_halos) == len(halos)

def test_autodetect_subfind(snap):
    assert isinstance(snap.halos(), pynbody.halo.subfind.SubfindCatalogue)

def test_autodetect_subfind_from_filename(snap):
    snap._filename = ""
    assert isinstance(snap.halos(filename="testdata/subfind/groups_019"), pynbody.halo.subfind.SubfindCatalogue)

def test_group_properties(groups):
    assert_allclose(float(groups[3].properties['mmean_200']), 1.22e14, rtol=1.e-2)
    assert_allclose(float(groups[3].properties['rtop_200']), 1.02, rtol=1.e-2)

@pytest.mark.parametrize('with_units', [True, False])
def test_all_group_properties(groups, with_units):
    properties = groups.get_properties_all_halos(with_units=with_units)
    assert_allclose(properties['mmean_200'][3], 12181.62, rtol=1.e-4)
    assert_allclose(properties['rtop_200'][3], 1.02, rtol=1.e-2)
    assert np.all(properties['children'][3] == np.arange(89,109))
    if with_units:
        assert properties['mmean_200'].units == '1e10 Msol h^-1'
        assert properties['rtop_200'].units == 'Mpc a h^-1'
    else:
        assert not hasattr(properties['mmean_200'], 'units')
        assert not hasattr(properties['rtop_200'], 'units')

def test_group_children(groups):
    assert np.all(groups[0].properties['children'] == np.arange(0, 39))
    assert np.all(groups[3].properties['children'] == np.arange(89,109))
    halo_nums_in_subs = [sh.properties['halo_number'] for sh in groups[3].subhalos]
    assert (halo_nums_in_subs == np.arange(89,109)).all()

    halos = groups.base.halos(subhalos=True)
    for sh_from_group, sh in zip(groups[3].subhalos, halos[89:109]):
        assert sh.properties['halo_number'] == sh_from_group.properties['halo_number']
        assert np.all(sh['iord'] == sh_from_group['iord'])

    # subs of subs also tested below in test_halo_children
    assert groups[1].subhalos[0].subhalos[0].properties['halo_number'] == 39

def test_halo_properties(halos):
    assert_allclose(halos[3].properties['sub_pos'], [ 0.4759067,  1.862322 , 33.249245 ])
    assert_allclose(float(halos[0].properties['sub_VMax']), 685., rtol=1.e-2)

@pytest.mark.parametrize('with_units', [True, False])
def test_all_halo_properties(halos, with_units):
    properties = halos.get_properties_all_halos(with_units=with_units)
    assert_allclose(properties['sub_pos'][3], [ 0.4759067,  1.862322 , 33.249245 ])
    assert_allclose(float(properties['sub_VMax'][0]), 685., rtol=1.e-2)
    for i in range(20):
        # NB this assumes the halo number = the halo offset, but that's ok for this file
        assert (properties['children'][i] == halos[i].properties['children']).all()
    if with_units:
        assert properties['sub_pos'].units == 'Mpc a h^-1'
        assert properties['sub_VMax'].units == 'km a**1/2 s^-1'

def test_halo_properties_physical(halos):
    halos.physical_units()
    assert_allclose(halos[3].properties['sub_pos'], [  558.2842,  2184.6821, 39004.55  ])

def test_halo_children(halos):
    assert np.all(halos[0].properties['children'] == [ 0,  1,  2,  3,  4,  6,  7, 12, 13, 14, 15, 16, 17, 20, 21, 22, 24,
        26, 32, 33, 35, 36, 37])
    assert np.all(halos[1].properties['children'] == [ 5,  8,  9, 11, 18, 19, 23, 25, 29, 30, 31, 34])
    assert halos[39].properties['sub_groupNr'] == 1
    assert np.all(halos[39].properties['children'] == [39, 40, 41, 42, 43, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56,
        58, 59, 61, 62, 63, 64, 65, 66, 69])

    for sh_a, expected_halo_num in zip(halos[39].subhalos,
                          [39, 40, 41, 42, 43, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56,
                           58, 59, 61, 62, 63, 64, 65, 66, 69]
                          ):
        sh_b = halos[expected_halo_num]
        assert sh_a.properties['halo_number'] == sh_b.properties['halo_number']
        assert np.all(sh_a['iord'] == sh_b['iord'])

def test_group_particles(groups):
    assert np.all(groups[0]['iord'][::1000] == [1902806, 2048342,   66771, 2083281, 2066768, 2000586,   35672,
          1966932, 1901524,   83913, 2035150, 1835347,   49759, 1836627,
            84691, 1900752,  132302, 2050529,   67936,  132548, 1950435,
          1936453, 2047847, 1851964,  214227, 1981505,  145348, 1883081,
          2095318, 1965514, 2079065,   46789, 2062554, 1914447,  228794,
          1932736, 2095170,  211516,   46020, 1914966, 1901016, 1933909,
            50884,  115538, 2028737,   81869,   79680, 1802047,  178240,
           146263])
    assert np.all(groups[10]['iord'][::1000] == [ 848941,  897584,  996144,  882349,  848173,  847673,  946356,
           733368,  997292, 1126063])


def test_halo_particles(halos):
    assert np.all(halos[0]['iord'][::1000] == [1902806, 2048342,   66771, 2083281, 2066768, 2000586,   35672,
          1966932, 1901524,   83913, 2035150, 1835347,   49759, 1836627,
            84691, 1900752,  132302, 2050529,   67936,  132548, 1950435,
          1936453, 2047847, 1851964,  214227])
    assert np.all(halos[39]['iord'][::1000] == [493394, 198751, 396116, 462038, 394458, 328930, 623070, 525789,
          280024, 623448, 245976, 608093, 608598, 494178, 491350, 507866,
          264776, 558562, 362319, 591560, 706259, 348265, 249821, 231233,
          474710, 231023, 380767, 413674, 589153, 459711, 576598, 671059,
          523615])



"""
Efficient 3D KDTree implementation for fast geometrical calculations such as neighbour lists and smoothing lengths.

"""
import logging
import time
import warnings

import numpy as np

from .. import array as ar, config, util
from . import kdmain

logger = logging.getLogger("pynbody.kdtree")

# Boundary type definition must exactly match the C++ definition (see kd.h)
Boundary = np.dtype([
    ('fMin', np.float32, (3,)),
    ('fMax', np.float32, (3,))
])

# KDNode type definition must exactly match the C++ definition (see kd.h)
KDNode = np.dtype([
    ('fSplit', np.float32),
    ('bnd', Boundary),
    ('iDim', np.int32),
    ('pLower', np.intp),
    ('pUpper', np.intp)
])

class KDTree:
    """KDTree can be used for smoothing, interpolating and geometrical queries.


    You are unlikely to need to interact with the KDTree directly, instead using the higher-level
    SPH functionality such as

    >>> f = pynbody.load("my_snapshot")
    >>> f.gas['smooth'] # will build KDTree and calculate smoothing lengths for gas, if not available on disk
    >>> f.gas['rho'] # will calculate densities for gas, if not available on disk
    >>> f.dm['rho'] # will calculate densities for dark matter, if not available on disk

    KDTree can also be used to accelerate geometrical queries on a snapshot, e.g.:

    >>> f = pynbody.load('my_snapshot')
    >>> f.build_kdtree()
    >>> f[pynbody.filt.Sphere('10 kpc')]

    See :ref:`performance of filters <filters_tutorial_performance_implications>` for more information.

    Most KDTree operations proceed in parallel, using the number of threads specified in the
    pynbody configuration.

    Performance statistics can be tested using the ``performance_kdtree.py`` script in the tests folder.

    Note that the KDTree takes into account periodicity of cosmological volumes if it can. However, as soon as a
    snapshot has been rotated, this may become impossible. In this case, the KDTree will issue a warning,
    and will not use periodicity. This is to avoid incorrect results due to the periodicity assumption,
    but it will lead to artefacts on the edge of the box. If you are working with a cosmological simulation,
    it is best to perform any KDTree operations (such as smoothing) before rotating the snapshot, e.g.:

    >>> f = pynbody.load('my_snapshot')
    >>> f['rho']; f['smooth'] # force KDTree to be built and density estimations made
    >>> pynbody.analysis.faceon(f) # rotate the snapshot
    >>> pynbody.plot.image(f.g, qty='rho', width='10 kpc')

    """

    PROPID_HSM = 1
    PROPID_RHO = 2
    PROPID_QTYMEAN_1D = 3
    PROPID_QTYMEAN_ND = 4
    PROPID_QTYDISP_1D = 5
    PROPID_QTYDISP_ND = 6
    PROPID_QTYDIV  = 7
    PROPID_QTYCURL = 8

    def __init__(self, pos, mass, leafsize=32, boxsize=None, num_threads=None, shared_mem=False):
        """Create a KDTree

        Most users will not need to call this directly, but will instead use the higher-level functions, either
        via accessing SPH properties on a snapshot, or calling :meth:`pynbody.snapshot.simsnap.SimSnap.build_tree`.

        Parameters
        ----------
        pos : pynbody.array.SimArray
            Particles positions.
        mass : pynbody.array.SimArray
            Particles mass.
        leafsize : int, optional
            The number of particles in leaf nodes (default 32).
        boxsize : float, optional
            Boxsize (default None)
        num_threads : int, optional
            Number of threads to use when building tree (if None, use configured/detected number of processors).
        shared_mem : bool, optional
            Whether to keep kdtree in shared memory so that it can be shared between processes (default False).
        """

        num_threads = self._set_num_threads(num_threads)

        # get a power of 2 for num_threads to pass to the constructor, because
        # otherwise the workload will not be balanced across threads and they
        # will be wasted
        num_threads_init = 2 ** int(np.log2(num_threads))


        self.leafsize = int(leafsize)
        self.kdtree = kdmain.init(pos, mass, self.leafsize)
        nodes = kdmain.get_node_count(self.kdtree)

        # zeroing the KDNode elements below isn't strictly necessary but helps with testing/debugging since some
        # of the memory won't be used

        if shared_mem:
            from ..array import shared
            self.kdnodes = shared.make_shared_array(nodes, KDNode, zeros=True)
            self.particle_offsets = shared.make_shared_array(len(pos), np.intp)
        else:
            self.kdnodes = np.zeros(nodes, dtype=KDNode)
            self.particle_offsets = np.empty(len(pos), dtype=np.intp)

        kdmain.build(self.kdtree, self.kdnodes, self.particle_offsets, num_threads_init)

        self.boxsize = boxsize
        self._pos = pos
        self.set_kernel(config['sph'].get('Kernel', 'CubicSpline'))

    def _set_num_threads(self, num_threads):
        if num_threads is None:
            num_threads = int(config["number_of_threads"])
        self.num_threads = num_threads
        return num_threads

    def serialize(self):
        """Produce a serialized description of the tree"""
        return self.leafsize, self.boxsize, self.kdnodes, self.particle_offsets, self._kernel_id

    @classmethod
    def deserialize(cls, pos, mass, serialized_state, num_threads = None, boxsize=None):
        """Reconstruct a KDTree from a serialized state."""
        self = object.__new__(cls)
        self._set_num_threads(num_threads)
        leafsize, boxsize_s, kdnodes, particle_offsets, kernel_id = serialized_state
        if boxsize is not None and abs(boxsize-boxsize_s) > 1e-6:
            warnings.warn("Boxsize in serialized state does not match boxsize passed to deserialize")
        self.kdtree = kdmain.init(pos, mass, leafsize)
        self.leafsize = leafsize
        nodes = kdmain.get_node_count(self.kdtree)
        if len(kdnodes) != nodes:
            raise ValueError("Number of nodes in serialized state does not match number of nodes in kdtree")
        self.kdnodes = kdnodes
        if len(particle_offsets) != len(pos):
            raise ValueError("Number of particle offsets in serialized state does not match number of particles")
        self.particle_offsets = particle_offsets
        self.boxsize = boxsize
        self._pos = pos
        self._kernel_id = kernel_id

        kdmain.import_prebuilt(self.kdtree, self.kdnodes, self.particle_offsets, 0)

        return self



    def particles_in_sphere(self, center, radius):
        """Find particles within a sphere.

        Most users will not need to call this directly, but will instead use :class:`pynbody.filt.Sphere`.

        Parameters
        ----------
        center : array_like
            Center of the sphere.
        radius : float
            Radius of the sphere.

        Returns
        -------
        indices : array_like
            Indices of the particles within the sphere.
        """
        smx = kdmain.nn_start(self.kdtree, 1, self.boxsize)

        particle_ids = kdmain.particles_in_sphere(self.kdtree, smx, center[0], center[1], center[2], radius)

        kdmain.nn_stop(self.kdtree, smx)
        return particle_ids

    def nn(self, nn=None):
        """Generator of neighbour list.

        Parameters
        ----------
        nn : int, None
            number of neighbours. If None, default to 64.

        Yields
        ------
        nbr_list : list[int, float, list[int], list[float]]

            Information about the neighbours of a particle.

            List with four elements:

            * ``nbr_list[0]`` (``int``): the index of the particle in the snapshot's arrays.
            * ``nbr_list[1]`` (``float``): the smoothing length of the particle
            * ``nbr_list[2]`` (``list[int]``): list of the indexes of the `nn` neighbouring particles.
            * ``nbr_list[3]`` (``list[float]``): list of distances squared of each neighbouring particles.

            The lists of particle and of the relative distance squared are not sorted.

        """
        if nn is None:
            nn = 64

        smx = kdmain.nn_start(self.kdtree, int(nn), self.boxsize)
        kdmain.domain_decomposition(self.kdtree, 1)

        while True:
            nbr_list = kdmain.nn_next(self.kdtree, smx)
            if nbr_list is None:
                break
            yield nbr_list

        kdmain.nn_stop(self.kdtree, smx)

    def all_nn(self, nn=None):
        """A list of neighbours information for all the particles in the snapshot."""
        return [x for x in self.nn(nn)]

    @staticmethod
    def array_name_to_id(name):
        """Convert the pynbody name of an array to the corresponding integer ID in the C++ code."""
        if name == "smooth":
            return 0
        elif name == "rho":
            return 1
        elif name == "mass":
            return 2
        elif name == "qty":
            return 3
        elif name == "qty_sm":
            return 4
        else:
            raise ValueError("Unknown KDTree array")

    def set_array_ref(self, name, ar):
        """Give the C++ code access to a given array.

        Parameters
        ----------

        name : str
            Name of the array to set (can be 'smooth', 'rho', 'mass', 'qty', 'qty_sm')
        ar : pynbody.array.SimArray
            Array that the C++ code will access. Note that INCREF is called on the array, so it will be kept
            alive as long as the KDTree object is alive (or until another set_array_ref call replaces it.)
        """

        if self.array_name_to_id(name) < 3:
            if not np.issubdtype(self._pos.dtype, ar.dtype):
                raise TypeError(
                    "KDTree requires matching dtypes for %s (%s) and pos (%s) arrays"
                    % (name, ar.dtype, self._pos.dtype)
                )
        kdmain.set_arrayref(self.kdtree, self.array_name_to_id(name), ar)
        assert self.get_array_ref(name) is ar

    def get_array_ref(self, name):
        """Get the current array reference for a given name ('smooth', 'rho', 'mass', 'qty', or 'qty_sm')."""
        return kdmain.get_arrayref(self.kdtree, self.array_name_to_id(name))

    def smooth_operation_to_id(self, name):
        """Convert the name of a smoothing operation to the corresponding integer ID in the C++ code.

        Valid names are:

        * 'hsm' : compute smoothing lengths
        * 'rho' : compute density
        * 'qty_mean' : compute SPH mean of the 'qty' array
        * 'qty_disp' : compute SPH dispersion of the 'qty' array
        * 'qty_div' : compute divergence of the 'qty' array
        * 'qty_curl' : compute curl of the 'qty' array

        """

        if name == "hsm":
            return self.PROPID_HSM
        elif name == "rho":
            return self.PROPID_RHO
        elif name == "qty_mean":
            input_array = self.get_array_ref("qty")
            if len(input_array.shape) == 1:
                return self.PROPID_QTYMEAN_1D
            elif len(input_array.shape) == 2:
                if input_array.shape[1] != 3:
                    raise ValueError("Currently only able to smooth 3D or 1D arrays")
                return self.PROPID_QTYMEAN_ND
        elif name == "qty_disp":
            input_array = self.get_array_ref("qty")
            if len(input_array.shape) == 1:
                return self.PROPID_QTYDISP_1D
            elif len(input_array.shape) == 2:
                if input_array.shape[1] != 3:
                    raise ValueError("Currently only able to smooth 3D or 1D arrays")
                return self.PROPID_QTYDISP_ND
        elif name == "qty_div":
            input_array = self.get_array_ref("qty")
            if len(input_array.shape) != 2 and input_array.shape[1] != 3:
                raise ValueError("Can only compute divergence of 3D arrays")
            return self.PROPID_QTYDIV
        elif name == "qty_curl":
            input_array = self.get_array_ref("qty")
            if len(input_array.shape) != 2 and input_array.shape[1] != 3:
                raise ValueError("Can only compute curl of 3D arrays")
            return self.PROPID_QTYCURL
        else:
            raise ValueError("Unknown smoothing request %s" % name)


    def set_kernel(self, kernel = 'CubicSpline'):
        """Set the kernel for smoothing operations.

        Parameters
        ----------
        kernel : str
            Keyword to specify the smoothing kernel. Options: 'CubicSpline', 'WendlandC2'
        """
        if kernel == 'CubicSpline':
            self._kernel_id = 0
        elif kernel == 'WendlandC2':
            self._kernel_id = 1
        else:
            raise ValueError("Kernel %r not recognised. Please choose either 'CubicSpline' or 'WendlandC2'." % kernel)


    def populate(self, mode, nn):
        """Create the KDTree and perform the operation specified by `mode`.

        Parameters
        --------
        mode : str (see `kdtree.smooth_operation_to_id`)
            Specify operation to perform (compute smoothing lengths, density, SPH mean, or SPH dispersion).
        nn : int
            Number of neighbours to be considered when smoothing.
        """


        if nn is None:
            nn = 64

        smx = kdmain.nn_start(self.kdtree, int(nn), self.boxsize)

        try:
            propid = self.smooth_operation_to_id(mode)

            if propid == self.PROPID_HSM:
                kdmain.domain_decomposition(self.kdtree, self.num_threads)


            if self.num_threads == 1:
                NT = 8
                util.thread_map(
                    kdmain.populate,
                    [self.kdtree] * NT,
                    [smx] * NT,
                    [propid] * NT,
                    list(range(0, NT)),
                    [self._kernel_id] * NT
                )
            else:
                util.thread_map(
                    kdmain.populate,
                    [self.kdtree] * self.num_threads,
                    [smx] * self.num_threads,
                    [propid] * self.num_threads,
                    list(range(0, self.num_threads)),
                    [self._kernel_id] * self.num_threads
                )
        finally:
            # Free C-structures memory
            kdmain.nn_stop(self.kdtree, smx)

    def sph_mean(self, array, nsmooth=64):
        r"""Calculate the SPH mean of a simulation array.

        Given a kernel W, this calculates

        .. math::

          r_i = \sum_{j=0}^{N_{smooth}}\left( \frac{m_j}{\rho_j} q_i W(|x_i - x_j|, s_i) \right)

        where r is the result of the smoothing, m is the mass array, rho is the density array, q is the input array,
        s_i is the smoothing length, and W is the kernel function.

        The actual computation is done in the C++-extension function smooth.cpp::smMeanQty[1,N]D.

        Parameters
        ----------
        array : pynbody.array.SimArray
            Quantity to smooth (compute SPH interpolation at particles position).

        nsmooth:
            Number of neighbours to use when smoothing.

        Returns
        -------
        output : pynbody.array.SimArray
            The SPH mean of the input array.
        """
        output = np.empty_like(array)

        if hasattr(array, "units"):
            output = output.view(ar.SimArray)
            output.units = array.units

        self.set_array_ref("qty", array)
        self.set_array_ref("qty_sm", output)

        logger.info("Smoothing array with %d nearest neighbours" % nsmooth)
        start = time.time()
        self.populate("qty_mean", nsmooth)
        end = time.time()

        logger.info("SPH smooth done in %5.3g s" % (end - start))

        return output

    def sph_dispersion(self, array, nsmooth=64):
        r"""Calculate the SPH dispersion of a simulation array.

        Given a kernel W, this routine computes the smoothed quantity:

        .. math::

          r_i = \sum_{j=0}^{N_{smooth}}\left( \frac{m_j}{\rho_j} q_i W(|x_i - x_j|, s_i) \right)

        Then the squared root of the SPH smoothed variance:

        .. math::

          d_i = \left\{ \sum_{j=0}^{N_{smooth}} \frac{m_j}{\rho_j} (q_i - r_i)^2 W(|x_i - x_j|, s_i) \right\}^{1/2}

        where d_i is the calculated dispersion, m is the mass array, rho is the density array, q is the input array,
        s_i is the smoothing length, and W is the kernel function.

        Actual computation is done in the C++-extension functions smooth.cpp::smDispQty[1,N]D.

        Parameters
        ----------
        array : pynbody.array.SimArray
            Quantity to compute dispersion of.

        nsmooth: int
            Number of neighbours to use when smoothing.

        Returns
        -------
        output : pynbody.array.SimArray
            The dispersion of the input array.
        """
        output = np.empty_like(array)
        if hasattr(array, "units"):
            output = output.view(ar.SimArray)
            output.units = array.units

        self.set_array_ref("qty", array)
        self.set_array_ref("qty_sm", output)

        logger.info("Getting dispersion of array with %d nearest neighbours" % nsmooth)
        start = time.time()
        self.populate("qty_disp", nsmooth)
        end = time.time()

        logger.info("SPH dispersion done in %5.3g s" % (end - start))

        return output

    def _sph_differential_operator(self, array, op, nsmooth=64):
        if op == "div":
            op_label = "divergence"
            output = np.empty(len(array), dtype=array.dtype)
        elif op == "curl":
            op_label = "curl"
            output = np.empty_like(array)
        else:
            raise ValueError("Can only compute curl or divergence")

        if hasattr(array, "units"):
            output = output.view(ar.SimArray)
            output.units = array.units/self._pos.units

        self.set_array_ref("qty", array)
        self.set_array_ref("qty_sm", output)

        logger.info("Getting %s of array with %d nearest neighbours" % (op_label, nsmooth))
        start = time.time()
        self.populate("qty_%s" % op, nsmooth)
        end = time.time()

        logger.info(f"SPH {op_label} done in {end - start:5.3g} s")

        return output

    def sph_curl(self, array, nsmooth=64):
        """Calculate the curl of a given array using SPH smoothing.

        Parameters
        ----------

        array : pynbody.array.SimArray
            Quantity to compute curl of.

        nsmooth : int
            Number of neighbours to use when smoothing.

        Returns
        -------
        pynbody.array.SimArray:
            The curl of the input array.
        """
        return self._sph_differential_operator(array, "curl", nsmooth)

    def sph_divergence(self, array, nsmooth=64):
        """Calculate the divergence of a given array using SPH smoothing.

        Parameters
        ----------

        array : pynbody.array.SimArray
            Quantity to compute curl of.

        nsmooth : int
            Number of neighbours to use when smoothing.

        Returns
        -------
        pynbody.array.SimArray:
            The curl of the input array.
        """
        return self._sph_differential_operator(array, "div", nsmooth)

    def __del__(self):
        if hasattr(self, "kdtree"):
            kdmain.free(self.kdtree)



"""
derived
=======

Holds procedures for creating new arrays from existing ones, e.g. for
getting the radial position. For more information see :ref:`derived`.

"""

import functools
import logging
import time
import warnings

import numpy as np

from . import analysis, array, config, units
from .dependencytracker import DependencyError
from .snapshot import SimSnap

logger = logging.getLogger('pynbody.derived')


@SimSnap.derived_array
def r(self):
    """Radial position"""
    return ((self['pos'] ** 2).sum(axis=1)) ** (1, 2)


@SimSnap.derived_array
def rxy(self):
    """Cylindrical radius in the x-y plane"""
    return ((self['pos'][:, 0:2] ** 2).sum(axis=1)) ** (1, 2)


@SimSnap.derived_array
def vr(self):
    """Radial velocity"""
    return (self['pos'] * self['vel']).sum(axis=1) / self['r']


@SimSnap.derived_array
def v2(self):
    """Squared velocity"""
    return (self['vel'] ** 2).sum(axis=1)


@SimSnap.derived_array
def vt(self):
    """Tangential velocity"""
    return np.sqrt(self['v2'] - self['vr'] ** 2)


@SimSnap.derived_array
def ke(self):
    """Specific kinetic energy"""
    return 0.5 * (self['vel'] ** 2).sum(axis=1)


@SimSnap.derived_array
def te(self):
    """Specific total energy"""
    return self['ke'] + self['phi']


@SimSnap.derived_array
def j(self):
    """Specific angular momentum"""
    angmom = np.cross(self['pos'], self['vel']).view(array.SimArray)
    angmom.units = self['pos'].units * self['vel'].units
    return angmom


@SimSnap.derived_array
def j2(self):
    """Square of the specific angular momentum"""
    return (self['j'] ** 2).sum(axis=1)


@SimSnap.derived_array
def jz(self):
    """z-component of the angular momentum"""
    return self['j'][:, 2]


@SimSnap.derived_array
def vrxy(self):
    """Cylindrical radial velocity in the x-y plane"""
    return (self['pos'][:, 0:2] * self['vel'][:, 0:2]).sum(axis=1) / self['rxy']


@SimSnap.derived_array
def vcxy(self):
    """Cylindrical tangential velocity in the x-y plane"""
    f = (self['x'] * self['vy'] - self['y'] * self['vx']) / self['rxy']
    f[np.where(f != f)] = 0
    return f


@SimSnap.derived_array
def vphi(self):
    """Azimuthal velocity (synonym for vcxy)"""
    return self['vcxy']


@SimSnap.derived_array
def vtheta(self):
    """Velocity projected to polar direction"""
    return (np.cos(self['az']) * np.cos(self['theta']) * self['vx'] +
            np.sin(self['az']) * np.cos(self['theta']) * self['vy'] -
            np.sin(self['theta']) * self['vz'])


_op_dict = {"mean": "mean velocity",
            "disp": "velocity dispersion",
            "curl": "velocity curl",
            "div": "velocity divergence",
            }


def _v_sph_operation(self, op):
    """SPH-smoothed velocity operations"""
    from . import sph

    self.build_tree()

    nsmooth = config['sph']['smooth-particles']

    logger.info('Calculating %s with %d nearest neighbours' % (_op_dict[op], nsmooth))

    if op in ['mean', 'curl']:
        sm = array.SimArray(np.empty_like(self['vel']), self['vel'].units)
    else:
        sm = array.SimArray(np.empty(len(self['vel']), dtype=self['vel'].dtype), self['vel'].units)

    if op in ['div', 'curl']:
        sm.units /= self['pos'].units

    self.kdtree.set_array_ref('rho', self['rho'])
    self.kdtree.set_array_ref('smooth', self['smooth'])
    self.kdtree.set_array_ref('mass', self['mass'])
    self.kdtree.set_array_ref('qty', self['vel'])
    self.kdtree.set_array_ref('qty_sm', sm)

    start = time.time()
    self.kdtree.populate('qty_%s' % op, nsmooth)
    end = time.time()

    logger.info(f'{_op_dict[op]} done in {end - start:5.3g} s')

    return sm


@SimSnap.derived_array
def v_mean(self):
    """SPH-smoothed mean velocity"""
    return _v_sph_operation(self, "mean")

@SimSnap.derived_array
def v_disp(self):
    """SPH-smoothed velocity dispersion"""
    return _v_sph_operation(self, "disp")

@SimSnap.derived_array
def v_curl(self):
    """SPH-smoothed curl of velocity"""
    return _v_sph_operation(self, "curl")

@SimSnap.derived_array
def vorticity(self):
    """SPH-smoothed vorticity"""
    return _v_sph_operation(self, "curl")

@SimSnap.derived_array
def v_div(self):
    """SPH-smoothed divergence of velocity"""
    return _v_sph_operation(self, "div")

@SimSnap.derived_array
def age(self):
    """Stellar age determined from formation time and current snapshot time"""
    return self.properties['time'].in_units(self['tform'].units, **self.conversion_context()) - self['tform']

bands_available = ['u', 'b', 'v', 'r', 'i', 'j', 'h', 'k', 'U', 'B', 'V', 'R', 'I',
                   'J', 'H', 'K']

def lum_den_template(band, s):
        val = (10 ** (-0.4 * s[band + "_mag"])) * s['rho'] / s['mass']
        val.units = s['rho'].units/s['mass'].units
        return val

for band in bands_available:
    X = lambda s, b=str(band): analysis.luminosity.calc_mags(s, band=b)
    X.__name__ = band + "_mag"
    X.__doc__ = band + " magnitude from analysis.luminosity.calc_mags"""
    SimSnap.derived_array(X)

    lum_den = functools.partial(lum_den_template,band)

    lum_den.__name__ = band + "_lum_den"
    lum_den.__doc__ = "Luminosity density in astronomy-friendly units: 10^(-0.4 %s_mag) per unit volume. " \
                      "" \
                      "The magnitude is taken from analysis.luminosity.calc_mags."%band
    SimSnap.derived_array(lum_den)


@SimSnap.derived_array
def theta(self):
    """Angle from the z axis, from [0:pi]"""
    return np.arccos(self['z'] / self['r'])


@SimSnap.derived_array
def alt(self):
    """Angle from the horizon, from [-pi/2:pi/2]"""
    return np.pi / 2 - self['theta']


@SimSnap.derived_array
def az(self):
    """Angle in the xy plane from the x axis, from [-pi:pi]"""
    return np.arctan2(self['y'], self['x'])


@SimSnap.derived_array
def cs(self):
    """Sound speed"""
    return np.sqrt(5.0 / 3.0 * units.k * self['temp'] / self['mu'] / units.m_p)



@SimSnap.derived_array
def mu(sim, t0=None, Y=0.245):
    """Mean molecular mass, i.e. the mean atomic mass per particle. Assumes primordial abundances."""
    try:
        x = _mu_from_electron_frac(sim, Y)
    except (KeyError, DependencyError):
        try:
            x = _mu_from_HI_HeI_HeII_HeIII(sim)
        except KeyError:
            x = _mu_from_temperature_threshold(sim, Y, t0)

    x.units = units.Unit("1")
    return x


def _mu_from_temperature_threshold(sim, Y, t0):
    warnings.warn("No ionization fractions found, assuming fully ionised gas above 10^4 and neutral below 10^4K. "
                  "This is a very crude approximation.")
    x = np.empty(len(sim)).view(array.SimArray)
    if t0 is None:
        t0 = sim['temp']
    x[np.where(t0 >= 1e4)[0]] = 4. / (8 - 5 * Y)
    x[np.where(t0 < 1e4)[0]] = 4. / (4 - 3 * Y)
    return x


def _mu_from_HI_HeI_HeII_HeIII(sim):
    x = sim["HI"] + 2 * sim["HII"] + sim["HeI"] + \
        2 * sim["HeII"] + 3 * sim["HeIII"]
    x = x ** -1
    return x

def _mu_from_electron_frac(sim, Y):
    return 4./(4.-3.*Y+4*(1.-Y)*sim['ElectronAbundance'])



@SimSnap.derived_array
def p(sim):
    """Pressure"""
    p = sim["u"] * sim["rho"] * (2. / 3)
    p.convert_units("Pa")
    return p


@SimSnap.derived_array
def u(self):
    """Gas internal energy derived from temperature"""
    gamma = 5. / 3
    return self['temp'] * units.k / (self['mu'] * units.m_p * (gamma - 1))


@SimSnap.derived_array
def temp(self):
    """Gas temperature derived from internal energy"""
    gamma = 5. / 3
    mu_est = np.ones(len(self))
    for i in range(5):
        temp = (self['u'] * units.m_p / units.k) * (mu_est * (gamma - 1))
        temp.sim = self # to allow use of conversion context, e.g. scalefactor
        temp.convert_units("K")
        mu_est = mu(self, temp)
    return temp


@SimSnap.derived_array
def zeldovich_offset(self):
    """The position offset in the current snapshot according to
    the Zel'dovich approximation applied to the current velocities.
    (Only useful in the generation or analysis of initial conditions.)"""
    from . import analysis
    bdot_by_b = analysis.cosmology.rate_linear_growth(
        self, unit='km Mpc^-1 s^-1') / analysis.cosmology.linear_growth_factor(self)

    a = self.properties['a']

    offset = self['vel'] / (a * bdot_by_b)
    offset.units = self['vel'].units / units.Unit('km Mpc^-1 s^-1 a^-1')
    return offset


@SimSnap.derived_array
def aform(self):
    """The expansion factor at the time specified by the tform array."""

    from . import analysis
    z = analysis.cosmology.redshift(self, self['tform'])
    a = 1. / (1. + z)
    return a

@SimSnap.derived_array
def tform(self):
    """The time of the specified expansion factor in the aform"""
    from . import analysis
    t = analysis.cosmology.age(self, 1./self['aform'] - 1.)
    return t

@SimSnap.derived_array
def iord_argsort(self):
    """Indices so that particles are ordered by increasing ids"""
    return np.argsort(self['iord'])



import numpy as np
import numpy.testing as npt
import pytest

import pynbody


def setup_module():
    global snap
    snap = pynbody.load("testdata/gadget2/test_g2_snap")


def teardown_module():
    global snap
    del snap


def test_construct():
    """Check the basic properties of the snapshot"""
    assert np.size(snap._files) == 2
    assert snap.header.num_files == 2
    assert snap.filename == "testdata/gadget2/test_g2_snap"
    assert snap._num_particles == 8192
    for f in snap._files:
        assert f.format2
        assert f.endian == "="

def test_properties():
    assert "time" in snap.properties


def test_loadable():
    """Check we have found all the blocks that should be in the snapshot"""
    blocks = snap.loadable_keys()
    expected_gas = ['nhp', 'smooth', 'nhe', 'u', 'sfr', 'pos',
                    'vel', 'iord', 'mass', 'nh', 'rho', 'nheq', 'nhep']
    expected_all = ['pos', 'vel', 'iord', 'mass']

    # Check that they have the right families
    assert(set(snap.gas.loadable_keys()) == set(expected_gas))
    assert(set(snap.dm.loadable_keys()) == set(expected_all))
    assert(set(snap.star.loadable_keys()) == set(expected_all))
    assert(set(snap.loadable_keys()) == set(expected_all))
    assert(snap.neutrino.loadable_keys() == [])


def test_standard_arrays():
    """Check we can actually load some of these arrays"""
    snap.dm['pos']
    snap.gas['pos']
    snap.star['pos']
    snap['pos']
    snap['mass']
    # Load a second time to check that family_arrays still work
    snap.dm['pos']
    snap['vel']
    snap['iord']
    snap.gas['rho']
    snap.gas['u']
    snap.star['mass']


def test_array_sizes():
    """Check we have the right sizes for the arrays"""
    assert(np.shape(snap.dm['pos']) == (4096, 3))
    assert(np.shape(snap['vel']) == (8192, 3))
    assert(np.shape(snap.gas['rho']) == (4039,))
    assert(snap.gas['u'].dtype == np.float32)
    assert(snap.gas['iord'].dtype == np.int32)


def test_fam_sim():
    """Check that an array loaded as families is the same as one loaded as a simulation array"""
    snap2 = pynbody.load("testdata/gadget2/test_g2_snap")
    snap3 = pynbody.load("testdata/gadget2/test_g2_snap")
    snap3.gas["pos"]
    snap3.dm["pos"]
    snap3.star["pos"]
    assert((snap3["pos"] == snap2["pos"]).all())


def test_array_contents():
    """Check some array elements"""
    assert(np.max(snap["iord"]) == 8192)
    assert(np.min(snap["iord"]) == 1)
    assert(np.mean(snap["iord"]) == 4096.5)

    # 10/11/13 - AP - suspect the following tests are incorrect
    # because ordering of file did not agree with pynbody ordering

    assert(abs(np.mean(snap["pos"]) - 1434.664) < 0.004)
    assert(abs(snap["pos"][52][1] - 456.69678) < 0.001)
    assert(abs(snap.gas["u"][100] - 438.39496) < 0.001)
    assert(abs(snap.dm["mass"][5] - 0.04061608) < 0.001)


def test_header():
    """Check some header properties"""
    assert(abs(snap.header.BoxSize - 3000.0) < 0.001)
    assert(abs(snap.header.HubbleParam - 0.710) < 0.001)
    assert(abs(snap.header.Omega0 - 0.2669) < 0.001)
    assert(snap.header.flag_cooling == 1)
    assert(snap.header.flag_metals == 0)


def test_g1_load():
    """Check we can load gadget-1 files also"""
    with pytest.warns(RuntimeWarning, match=r"Run out of block names in the config file. Using fallbacks: UNK\*"):
        snap2 = pynbody.load("testdata/gadget1.snap")


def test_write():
    """Check that we can write a new snapshot and read it again,
    and the written and the read are the same."""
    snap.write(filename='testdata/test_gadget_write')
    snap3 = pynbody.load('testdata/test_gadget_write')
    assert set(snap.loadable_keys()) == set(snap3.loadable_keys())
    npt.assert_equal(snap3["pos"].view(np.ndarray), snap["pos"])
    npt.assert_equal(snap3.gas["rho"].view(np.ndarray), snap.gas["rho"])
    assert snap3.check_headers(snap.header, snap3.header)


def test_conversion():
    """Check that we can convert a file from tipsy format and load it again"""
    snap4 = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")
    snap4.write(fmt=pynbody.snapshot.gadget.GadgetSnap,
                filename="testdata/test_conversion.gadget")
    snap5 = pynbody.load("testdata/test_conversion.gadget")


def test_write_single_array():
    """Check that we can write a single array and read it back"""
    snap["pos"].write(overwrite=True)
    snap6 = pynbody.load("testdata/gadget2/test_g2_snap")
    assert((snap6["pos"] == snap["pos"]).all())


def test_no_mass_block():
    f = pynbody.load("testdata/gadget_no_mass")
    f['mass']  # should succeed


def test_unit_persistence():
    f = pynbody.load("testdata/gadget2/test_g2_snap")

    # f2 is the comparison case - just load the whole
    # position array and convert it, simple
    f2 = pynbody.load("testdata/gadget2/test_g2_snap")
    f2['pos']
    f2.physical_units()

    f.gas['pos']
    f.physical_units()
    assert (f.gas['pos'] == f2.gas['pos']).all()

    # the following lazy-loads should lead to the data being
    # auto-converted
    f.dm['pos']
    assert (f.gas['pos'] == f2.gas['pos']).all()
    assert (f.dm['pos'] == f2.dm['pos']).all()

    # the final one is the tricky one because this will trigger
    # an array promotion and hence internally inconsistent units
    f.star['pos']

    assert (f.star['pos'] == f2.star['pos']).all()

    # also check it hasn't messed up the other bits of the array!
    assert (f.gas['pos'] == f2.gas['pos']).all()
    assert (f.dm['pos'] == f2.dm['pos']).all()

    assert (f['pos'] == f2['pos']).all()


def test_per_particle_loading():
    """Tests that loading one family at a time results in the
    same final array as loading all at once. There are a number of
    subtelties in the gadget handler that could mess this up by loading
    the wrong data."""

    f_all = pynbody.load("testdata/gadget2/test_g2_snap")
    f_part = pynbody.load("testdata/gadget2/test_g2_snap")

    f_part.dm['pos']
    f_part.star['pos']
    f_part.gas['pos']

    assert (f_all['pos'] == f_part['pos']).all()

def test_issue321():
    """L-PICOLA outputs single-precision with no mass block, which causes problems
    with testing kd-trees"""

    f = pynbody.load("testdata/lpicola/lpicola_z0p000.0")
    assert f['pos'].dtype==np.dtype('float32')
    assert f['mass'].dtype==np.dtype('float32')


def test_units_override():
    f = pynbody.load("testdata/gadget2/test_g2_snap.0")
    assert f.filename == "testdata/gadget2/test_g2_snap"
    assert f['pos'].units == "kpc a h^-1"

    # In this case the unit override system is not effective because
    # the final ".1" is not stripped away in the filename:
    # the file `gadget2/test_g2_snap.units` is not used
    f_no_unit_override = pynbody.load("testdata/gadget2/test_g2_snap.1")
    assert f_no_unit_override.filename == "testdata/gadget2/test_g2_snap.1"
    assert f_no_unit_override['pos'].units == "Mpc a h^-1"  # from default_config.ini


def test_ignore_cosmology():
    f = pynbody.load("testdata/gadget2/test_g2_snap.1")
    f.physical_units()
    np.testing.assert_allclose(f.properties['time'].in_units('Gyr'), 2.57689526)
    f_no_cosmo = pynbody.load("testdata/gadget2/test_g2_snap.1", ignore_cosmo=True)
    np.testing.assert_allclose(f_no_cosmo.properties['time'].in_units('Gyr'), 271.608952)



from __future__ import annotations

import os
import tempfile
from typing import Iterable

import h5py


class HdfVdsMaker:
    def __init__(self, hdf_files: list[h5py.File | str]):
        self._files = []
        for f in hdf_files:
            if isinstance(f, h5py.File):
                self._files.append(f)
            else:
                self._files.append(h5py.File(f, 'r'))

    def concatenation_keys(self) -> Iterable[str]:
        """Returns all keys to concatenate as VDS"""
        # TODO - make this general
        return['Subhalos', 'SubhaloParticles', 'NestedSubhalos']


    def copy_keys(self) -> Iterable[str]:
        """Returns all keys to copy from the first file into the VDS.

        Examples of copy keys are headers or one-off arrays"""
        # TODO - make this general
        return ['SnapshotId']

    def make_hdf_vfile(self, filepath: str) -> h5py.File:
        with h5py.File(name=filepath, mode='w') as hdf_vfile:
            for k in self.concatenation_keys():
                self.write_single_vds(k, hdf_vfile)

            for k in self.copy_keys():
                self.write_single_vds(k, hdf_vfile, first_only=True)

        return hdf_vfile

    def write_single_vds(self, key: str, target_hdf_file: h5py.File, first_only: bool=False):
        shape = None
        sources = []
        slices = []
        dtype = None
        offset = 0

        files = [self._files[0]] if first_only else self._files

        for f in files:
            source_dataset: h5py.Dataset = f[key]
            if shape is None:
                dtype = source_dataset.dtype
                shape = source_dataset.shape
            else:
                if dtype != source_dataset.dtype:
                    raise ValueError(f"The dtypes of array {key} are inconsistent between files")
                if source_dataset.shape[1:] != shape[1:]:
                    raise ValueError(f"The shapes of array {key} are inconsistent between files")
                shape = (source_dataset.shape[0] + shape[0],)+shape[1:]
            sources.append(h5py.VirtualSource(source_dataset))
            slices.append(slice(offset, offset+len(source_dataset)))
            offset+=len(source_dataset)

        layout = h5py.VirtualLayout(shape=shape, dtype=dtype)
        for slice_, vsource in zip(slices, sources):
            layout[slice_] = vsource

        target_hdf_file.create_virtual_dataset(key, layout)


    def get_temporary_hdf_vfile(self) -> h5py.File:
        """Create the HDF file with virtual datasets in a temporary directory, such that it is deleted on closure"""

        with tempfile.TemporaryDirectory() as tmpdirname:
            filepath = os.path.join(tmpdirname, "nofile.hdf5")
            # ideally one would simply use  backing_store=False, to File but then there doesn't seem to be a way
            # to actually use the file (the VDS views just returns zeros).
            # Instead we write then re-read it, which presumably carries minimal overhead but is a bit ugly.

            self.make_hdf_vfile(filepath)

            hdf_vfile = h5py.File(name=filepath, mode='r')
        # on exiting the with block, the temporary directory and file are unlinked, but the file won't actually be
        # erased until it's closed.
        return hdf_vfile



"""Analysis involving angular momentum.

"""
import logging

import numpy as np

from .. import array, filt, transformation, units, util
from . import halo

logger = logging.getLogger('pynbody.analysis.angmom')


def ang_mom_vec(snap):
    """
    Calculates the angular momentum vector of the specified snapshot.

    Parameters
    ----------

    snap : SimSnap
        The snapshot to analyze

    Returns
    -------

    array.SimArray
        The angular momentum vector of the snapshot

    """

    angmom = (snap['mass'].reshape((len(snap), 1)) *
              np.cross(snap['pos'], snap['vel'])).sum(axis=0).view(np.ndarray)
    result = angmom.view(array.SimArray)
    result.units = snap['mass'].units * snap['pos'].units * snap['vel'].units
    return result


@util.deprecated("ang_mom_vec_units is deprecated. Use ang_mom_vec instead.")
def ang_mom_vec_units(snap):
    """Deprecated alias for ang_mom_vec"""

    return ang_mom_vec(snap)


def spin_parameter(snap):
    """Return the spin parameter lambda' of a centered halo

    The spin parameter is defined as in eq. (5) of Bullock et al. 2001 (2001MNRAS.321..559B).

    Note that the halo has to be aligned such that its center is the coordinate
    origin and the velocity must be zeroed. If you are not sure whether this will
    be true, calculate the spin parameter of h using:

    >>> with pynbody.analysis.angmom.faceon(h):
    >>>     spin = pynbody.analysis.angmom.spin_parameter(h)


    Parameters
    ----------

    snap : SimSnap
        The snapshot to analyze

    Returns
    -------

    float
        The dimensionless spin parameter lambda' of the halo

    """

    m3 = snap['mass'].sum()
    m3 = m3 * m3 * m3
    l = np.sqrt(((ang_mom_vec_units(snap) ** 2).sum()) /
                (2 * units.G * m3 * snap['r'].max()))
    return float(l.in_units('1', **snap.conversion_context()))


def calc_sideon_matrix(angmom_vec, along = [1.0, 0.0, 0.0]):
    """Calculate the rotation matrix to put the specified angular momentum vector side-on.

    The rotation matrix is calculated such that the angular momentum vector will be placed in the y direction
    post-transformation.

    Parameters
    ----------
    angmom_vec : array_like
        The angular momentum vector that will be placed in the y-direction post-transformation.

    along : array_like
        An additional orientation vector. The components of this vector perpendicular to angmom_vec defines the
        direction to transform to 'along', i.e. to the positive x-axis post-transformation.

    Returns
    -------

    array_like
        The rotation matrix
    """
    vec_in = np.asarray(angmom_vec)
    vec_in = vec_in / np.sum(vec_in ** 2).sum() ** 0.5
    vec_p1 = np.cross(along, vec_in)
    vec_p1 = vec_p1 / np.sum(vec_p1 ** 2).sum() ** 0.5
    vec_p2 = np.cross(vec_in, vec_p1)

    matr = np.concatenate((vec_p2, vec_in, vec_p1)).reshape((3, 3))

    return matr


def calc_faceon_matrix(angmom_vec, up=[0.0, 1.0, 0.0]):
    """Calculate the rotation matrix to put the specified angular momentum vector face-on.

    The rotation matrix is calculated such that the angular momentum vector will be placed in the z direction
    post-transformation.

    Parameters
    ----------
    angmom_vec : array_like
        The angular momentum vector that will be placed in the z-direction post-transformation.
    up : array_like
        An additional orientation vector. The components of this vector perpendicular to angmom_vec defines the
        direction to transform to 'up', i.e. to the positive y-axis post-transformation.

    Returns
    -------

    array_like
        The rotation matrix

    """
    vec_in = np.asarray(angmom_vec)
    vec_in = vec_in / np.sum(vec_in ** 2).sum() ** 0.5
    vec_p1 = np.cross(up, vec_in)
    vec_p1 = vec_p1 / np.sum(vec_p1 ** 2).sum() ** 0.5
    vec_p2 = np.cross(vec_in, vec_p1)

    matr = np.concatenate((vec_p1, vec_p2, vec_in)).reshape((3, 3))

    return matr


def align(h, vec_to_xform, disk_size="5 kpc", move_all=True, already_centered = False,
           center_kwargs = None):
    """Reposition and rotate the ancestor of h to place the angular momentum into a specified orientation.

    The routine first calls the center routine to reposition the halo (unless already_centered is True).
    If there are a sufficient number of gas particles (more than 100), only the gas particles are used for
    centering, since these will also be used for angular momentum calculations; if there is an offset between
    e.g. dark matter and baryons, it is better to centre on the baryons.

    Then, it determines the disk orientation using the angular momentum vector of the gas particles within
    a specified radius of the halo center. If there is no gas within this radius, the routine falls back first
    to stellar particles, and then to all particles.

    Finally, the angular momentum vector is converted into a rotation matrix using the vec_to_xform function,
    and the rotation is applied.

    Parameters
    ----------

    h : SimSnap
      The portion of the simulation from which to extract a centre and orientation. Typically this is a galaxy halo.

    vec_to_xform : function
        The function to use to calculate the rotation matrix from the measured angular momentum vector.

    disk_size : str | float, optional
        The size of the disk to use for calculating the angular momentum vector. Default is "5 kpc".

    move_all : bool, optional
        If True, the ancestor simulation of *h* is transformed. If False, only *h* is moved. Default is True.

    already_centered : bool, optional
        If True, the simulation is assumed to be already centered. Default is False.

    center_kwargs : dict, optional
        Dictionary of additional keyword arguments to pass to the centering routine.

    """


    if center_kwargs is None:
        center_kwargs = {}

    center_kwargs.update({'move_all': move_all})

    if already_centered:
        if move_all:
            top = h.ancestor
        else:
            top = h
        tx = transformation.NullTransformation(top)
    else:
        if len(h.st) > 100:
            h_for_centering = h.st
        else:
            h_for_centering = h
        tx = halo.center(h_for_centering, **center_kwargs)

    try:
        if len(h.gas) > 5:
            cen = h.gas[filt.Sphere(disk_size)]
        elif len(h.st) > 5:
            cen = h.st[filt.Sphere(disk_size)]
        else:
            cen = h[filt.Sphere(disk_size)]

        logger.info("Calculating angular momentum vector...")
        trans = vec_to_xform(ang_mom_vec(cen))

        logger.info("Transforming simulation...")

        tx = tx.rotate(trans)

        logger.info("...done!")

        return tx

    except:
        tx.revert()


def sideon(h, **kwargs):
    """Reposition and rotate the ancestor of h to place the disk edge-on (i.e. into the x-z plane).

    Since pynbody's imaging routines project along the z direction, one can get a side-on view of a disk or
    other rotationally-supported structure by calling this routine first.

    For details of how the transformation is calculated, see the documentation for the underlying :func:`align` routine.

    Parameters
    ----------

    h : SimSnap
      The portion of the simulation from which to extract a centre and orientation. Typically this is a galaxy halo.

    disk_size : str | float, optional
        The size of the disk to use for calculating the angular momentum vector. Default is "5 kpc".

    move_all : bool, optional
        If True, the ancestor simulation of *h* is transformed. If False, only *h* is moved. Default is True.

    already_centered : bool, optional
        If True, the simulation is assumed to be already centered. Default is False.

    center_kwargs : dict, optional
        Dictionary of additional keyword arguments to pass to the centering routine.

    """

    kwargs.update({'vec_to_xform': calc_sideon_matrix})

    return align(h, **kwargs)


def faceon(h, **kwargs):
    """Reposition and rotate the ancestor of h to place the disk face-on (i.e. into the x-y plane).

    Since pynbody's imaging routines project along the z direction, one can get a face-on view of a disk
    or other rotationally-supported structure by calling this routine first.

    For details of how the transformation is calculated, see the documentation for the underlying :func:`align` routine.


    Parameters
    ----------

    h : SimSnap
      The portion of the simulation from which to extract a centre and orientation. Typically this is a galaxy halo.

    disk_size : str | float, optional
        The size of the disk to use for calculating the angular momentum vector. Default is "5 kpc".

    move_all : bool, optional
        If True, the ancestor simulation of *h* is transformed. If False, only *h* is moved. Default is True.

    already_centered : bool, optional
        If True, the simulation is assumed to be already centered. Default is False.

    center_kwargs : dict, optional
        Dictionary of additional keyword arguments to pass to the centering routine.

    """

    kwargs.update({'vec_to_xform': calc_faceon_matrix})

    return align(h, **kwargs)



from __future__ import annotations

import abc

import numpy as np

from pynbody.util import binary_search, is_sorted


class IordToOffset(abc.ABC):
    @abc.abstractmethod
    def map_ignoring_order(self, i: np.ndarray | int) -> np.ndarray | int:
        """Given an array of iord values, return the corresponding fpos values.

        Warning: The returned values are not guaranteed to be in the same order as the input iord array."""
        pass


class IordToOffsetDense(IordToOffset):
    def __init__(self, iord_array, max_iord=None):
        if max_iord is None:
            max_iord = int(iord_array.max())
        self._iord_to_offset = np.empty(max_iord + 1, dtype=np.int64)
        self._iord_to_offset.fill(-1)
        self._iord_to_offset[iord_array] = np.arange(len(iord_array), dtype=np.int64)

    def map_ignoring_order(self, i):
        return self._iord_to_offset[i]


class IordToOffsetSparse(IordToOffset):
    """Class for efficiently mapping from iords to offsets in the iord array, even if iord values are large.

    WARNING: if a query is made with iords that are not themselves in ascending order, a sort takes place
    ahead of the query and therefore the set returned is correct but the ordering is not preserved."""
    def __init__(self, iord_array):
        self._iord = iord_array
        self._iord_argsort = np.argsort(iord_array)

    def map_ignoring_order(self, iord_values: np.ndarray | int) -> np.ndarray | int:
        if not hasattr(iord_values, "__len__"):
            iord_values = np.array([iord_values])
            singleton = True
        else:
            iord_values = np.asarray(iord_values)
            singleton = False

            if is_sorted(iord_values) != 1:
                iord_values = np.sort(iord_values)

        result = binary_search(np.asarray(iord_values), self._iord, self._iord_argsort)

        if singleton:
            return result[0]
        else:
            return result

class IordOffsetModifier(IordToOffset):
    """A wrapper around an IordToOffset which adds a constant offset to the result of the underlying mapping.

    Useful if the iord values e.g. are only available for a single family; then the fpos_offset will correspond
    to the first index of that family in the pynbody snapshot.
    """
    def __init__(self, iord_to_offset: IordToOffset, fpos_offset: int):
        self._underlying = iord_to_offset
        self._fpos_offset = fpos_offset

    def map_ignoring_order(self, i: np.ndarray | int) -> np.ndarray | int:
        result = self._underlying.map_ignoring_order(i)
        result += self._fpos_offset
        return result

def make_iord_to_offset_mapper(iord: np.ndarray) -> IordToOffset:
    """Given an array of unique integers, iord, make an object which maps from an iord value to offset in the array.

    i.e. given an iord array and a subset of values my_iord_values,

     make_iord_to_offset_mapper(iord).map_ignoring_order(my_iord_values)

    returns the indexes of my_iord_values in the iord array.
    """

    max_iord = int(iord.max())
    assert iord.min() >= 0, "Can't handle negative iord values"

    if max_iord < 2 * len(iord):
        # maximum iord is not very big, just do a direct in-memory mapping for speed
        return IordToOffsetDense(iord, max_iord)
    else:
        # maximum iord is large, so we'll use util.binary_search to save memory at the cost of speed
        return IordToOffsetSparse(iord)



from typing import Any

import numpy as np
from numpy import typing as npt


class HaloParticleIndices:
    def __init__(self, particle_ids: npt.NDArray[int] = None, boundaries: np.ndarray[(Any, 2), int] = None):
        """An IndexList represents abstract information about halo membership

        * particle_ids: array of particle IDs (mutually exclusive with halo_number_per_particle), length Npart_in_halos
        * boundaries: a Nhalo x 2 array of start and stop indices for each halo in the particle_ids array

        NB throughout this class halo indices (zero-based, continuous integer numbers) are used, NOT halo numbers.
        For accessing halos by halo number, one must additionally use the HaloNumberMapper class to get the index
        before passing it in here.
        """

        self.particle_index_list = particle_ids
        self.particle_index_list_boundaries = boundaries


    def get_particle_index_list_for_halo(self, halo_index):
        """Get the index list for the specified halo index"""
        return self.particle_index_list[self._get_index_slice_for_halo(halo_index)]

    def _get_index_slice_for_halo(self, obj_offset):
        """Get the slice for the index array corresponding to the object *offset* (not ID),
        i.e. the one whose index list starts at self.boundaries[obj_offset]"""
        ptcl_start, ptcl_end = self.particle_index_list_boundaries[obj_offset]
        return slice(ptcl_start, ptcl_end)

    def get_halo_number_per_particle(self, sim_length, number_mapper, fill_value=-1, dtype=int):
        """Return an array of halo numbers, one per particle.

        Requires a HaloNumberMapper to map halo indices to halo numbers. If None is passed for the number_mapper,
        the halo indices are returned instead."""
        lengths = np.diff(self.particle_index_list_boundaries, axis=1).ravel()
        ordering = np.argsort(-lengths, kind='stable')

        id_array = np.empty(sim_length, dtype=dtype)
        id_array.fill(fill_value)

        if number_mapper is not None:
            halo_numbers = number_mapper.index_to_number(ordering)
        else:
            halo_numbers = range(len(ordering))

        for halo_number, halo_index in zip(halo_numbers, ordering):
            indexing_slice = self._get_index_slice_for_halo(halo_index)
            id_array[self.particle_index_list[indexing_slice]] = halo_number

        return id_array

    def __len__(self):
        return len(self.particle_index_list_boundaries)
